# -*- coding: utf-8 -*-

import json
from typing import Optional, Iterable, List, Union, Dict

from mdxpy import MdxBuilder, MdxHierarchySet, Member

from TM1py.Objects.Axis import ViewAxisSelection, ViewTitleSelection
from TM1py.Objects.Subset import Subset, AnonymousSubset
from TM1py.Objects.View import View
from TM1py.Utils import case_and_space_insensitive_equals


class NativeView(View):
    """ Abstraction of TM1 NativeView (classic cube view)

        :Notes:
            Complete, functional and tested
    """

    def __init__(self,
                 cube_name: str,
                 view_name: str,
                 suppress_empty_columns: Optional[bool] = False,
                 suppress_empty_rows: Optional[bool] = False,
                 format_string: Optional[str] = "0.#########",
                 titles: Optional[Iterable[ViewTitleSelection]] = None,
                 columns: Optional[Iterable[ViewAxisSelection]] = None,
                 rows: Optional[Iterable[ViewAxisSelection]] = None):
        super().__init__(cube_name, view_name)
        self._suppress_empty_columns = suppress_empty_columns
        self._suppress_empty_rows = suppress_empty_rows
        self._format_string = format_string
        self._titles = list(titles) if titles else []
        self._columns = list(columns) if columns else []
        self._rows = list(rows) if rows else []

    @property
    def body(self) -> str:
        return self._construct_body()

    @property
    def rows(self) -> List[ViewAxisSelection]:
        return self._rows

    @property
    def columns(self) -> List[ViewAxisSelection]:
        return self._columns

    @property
    def titles(self) -> List[ViewTitleSelection]:
        return self._titles

    @property
    def mdx(self):
        return self.as_MDX

    @property
    def MDX(self) -> str:
        return self.as_MDX

    @property
    def as_MDX(self) -> str:
        """ Build a valid MDX Query from an Existing cubeview. 
        Takes Zero suppression into account. 
        Throws an Exception when no elements are place on the columns.
        Subsets are referenced in the result-MDX through the TM1SubsetToSet Function

        :return: String, the MDX Query
        """
        if not self.columns:
            raise ValueError("Column selection must not be empty")

        query = MdxBuilder.from_cube(self.cube)
        if self._suppress_empty_rows:
            query.rows_non_empty()

        if self.suppress_empty_columns:
            query.columns_non_empty()

        axes = [self.columns]
        if self.rows:
            axes.append(self.rows)

        for axis_id, axis in enumerate(axes):
            for axis_selection in axis:
                subset = axis_selection.subset

                if isinstance(subset, AnonymousSubset):
                    if subset.expression is not None:
                        mdx_hierarchy_set = MdxHierarchySet.from_str(
                            dimension=subset.dimension_name,
                            hierarchy=subset.hierarchy_name,
                            mdx=subset.expression)

                    else:
                        members = [Member.of(subset.dimension_name, element) for element in subset.elements]
                        mdx_hierarchy_set = MdxHierarchySet.members(members)

                else:
                    mdx_hierarchy_set = MdxHierarchySet.tm1_subset_to_set(
                        dimension=axis_selection.dimension_name,
                        hierarchy=axis_selection.hierarchy_name,
                        subset=subset.name)
                query.add_hierarchy_set_to_axis(axis=axis_id, mdx_hierarchy_set=mdx_hierarchy_set)

        for title in self._titles:
            query.add_member_to_where(Member.of(title.dimension_name, title.selected))

        return query.to_mdx()

    @property
    def suppress_empty_cells(self) -> bool:
        return self._suppress_empty_columns and self._suppress_empty_rows

    @property
    def suppress_empty_columns(self) -> bool:
        return self._suppress_empty_columns

    @property
    def suppress_empty_rows(self) -> bool:
        return self._suppress_empty_rows

    @property
    def format_string(self) -> str:
        return self._format_string

    @suppress_empty_cells.setter
    def suppress_empty_cells(self, value: bool):
        self.suppress_empty_columns = value
        self.suppress_empty_rows = value

    @suppress_empty_rows.setter
    def suppress_empty_rows(self, value: bool):
        self._suppress_empty_rows = value

    @suppress_empty_columns.setter
    def suppress_empty_columns(self, value: bool):
        self._suppress_empty_columns = value

    @format_string.setter
    def format_string(self, value: str):
        self._format_string = value

    def add_column(self, dimension_name: str, subset: Union[Subset, AnonymousSubset] = None):
        """ Add Dimension or Subset to the column-axis

        :param dimension_name: name of the dimension
        :param subset: instance of TM1py.Subset. Can be None
        :return:
        """
        view_axis_selection = ViewAxisSelection(dimension_name=dimension_name, subset=subset)
        self._columns.append(view_axis_selection)

    def remove_column(self, dimension_name: str):
        """ remove dimension from the column axis

        :param dimension_name:
        :return:
        """
        for column in self._columns[:]:
            if case_and_space_insensitive_equals(column.dimension_name, dimension_name):
                self._columns.remove(column)

    def add_row(self, dimension_name: str, subset: Subset = None):
        """ Add Dimension or Subset to the row-axis

        :param dimension_name:
        :param subset: instance of TM1py.Subset. Can be None instead.
        :return:
        """
        view_axis_selection = ViewAxisSelection(dimension_name=dimension_name, subset=subset)
        self._rows.append(view_axis_selection)

    def remove_row(self, dimension_name: str):
        """ remove dimension from the row axis

        :param dimension_name:
        :return:
        """
        for row in self._rows[:]:
            if case_and_space_insensitive_equals(row.dimension_name, dimension_name):
                self._rows.remove(row)

    def add_title(self, dimension_name: str, selection: str, subset: Union[Subset, AnonymousSubset] = None):
        """ Add subset and element to the titles-axis

        :param dimension_name: name of the dimension.
        :param selection: name of an element.
        :param subset:  instance of TM1py.Subset. Can be None instead.
        :return:
        """
        view_title_selection = ViewTitleSelection(dimension_name, subset, selection)
        self._titles.append(view_title_selection)

    def remove_title(self, dimension_name: str):
        """ Remove dimension from the titles-axis

        :param dimension_name: name of the dimension.
        :return:
        """
        for title in self._titles[:]:
            if case_and_space_insensitive_equals(title.dimension_name, dimension_name):
                self._titles.remove(title)

    def substitute_title(self, dimension: str, element: str):
        for title in self._titles:
            if case_and_space_insensitive_equals(title.dimension_name, dimension):
                title._subset = AnonymousSubset(dimension, dimension, elements=[element])
                title._selected = element
                return

        raise ValueError(f"Dimension '{dimension}' not found in titles")

    @classmethod
    def from_json(cls, view_as_json: str, cube_name: Optional[str] = None) -> 'NativeView':
        """ Alternative constructor
                :Parameters:
                    `view_as_json` : string, JSON

                :Returns:
                    `View` : an instance of this class
        """
        view_as_dict = json.loads(view_as_json)
        return NativeView.from_dict(view_as_dict, cube_name)

    @classmethod
    def from_dict(cls, view_as_dict: Dict, cube_name: str = None) -> 'NativeView':
        titles, columns, rows = [], [], []

        for selection in view_as_dict['Titles']:
            if selection['Subset']['Name'] == '':
                subset = AnonymousSubset.from_dict(selection['Subset'])
            else:
                subset = Subset.from_dict(selection['Subset'])
            selected = selection['Selected']['Name'] if selection['Selected'] else ""
            titles.append(ViewTitleSelection(dimension_name=subset.dimension_name,
                                             subset=subset, selected=selected))
        for i, axe in enumerate([view_as_dict['Columns'], view_as_dict['Rows']]):
            for selection in axe:
                if selection['Subset']['Name'] == '':
                    subset = AnonymousSubset.from_dict(selection['Subset'])
                else:
                    subset = Subset.from_dict(selection['Subset'])
                axis_selection = ViewAxisSelection(dimension_name=subset.dimension_name,
                                                   subset=subset)
                columns.append(axis_selection) if i == 0 else rows.append(axis_selection)

        return cls(
            cube_name=view_as_dict["@odata.context"][20:view_as_dict["@odata.context"].find("')/")]
            if not cube_name else cube_name,
            view_name=view_as_dict['Name'],
            suppress_empty_columns=view_as_dict['SuppressEmptyColumns'],
            suppress_empty_rows=view_as_dict['SuppressEmptyRows'],
            format_string=view_as_dict['FormatString'],
            titles=titles,
            columns=columns,
            rows=rows)

    def _construct_body(self) -> str:
        """ construct the ODATA conform JSON representation for the NativeView entity.

        :return: string, the valid JSON
        """
        top_json = "{\"@odata.type\": \"ibm.tm1.api.v1.NativeView\",\"Name\": \"" + self._name + "\","
        columns_json = ','.join([column.body for column in self._columns])
        rows_json = ','.join([row.body for row in self._rows])
        titles_json = ','.join([title.body for title in self._titles])
        bottom_json = "\"SuppressEmptyColumns\": " + str(self._suppress_empty_columns).lower() + \
                      ",\"SuppressEmptyRows\":" + str(self._suppress_empty_rows).lower() + \
                      ",\"FormatString\": \"" + self._format_string + "\"}"
        return "".join([
            top_json,
            '\"Columns\":[',
            columns_json,
            '],\"Rows\":[',
            rows_json,
            '],\"Titles\":[',
            titles_json,
            '],',
            bottom_json])



from setuptools import setup

SCHEDULE_VERSION = '1.10.0'
SCHEDULE_DOWNLOAD_URL = (
        'https://github.com/Cubewise-code/TM1py/tarball/' + SCHEDULE_VERSION
)

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='TM1py',
    packages=['TM1py', 'TM1py/Exceptions', 'TM1py/Objects', 'TM1py/Services', 'TM1py/Utils'],
    version=SCHEDULE_VERSION,
    description='A python module for TM1.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    author='Marius Wirtz',
    author_email='MWirtz@cubewise.com',
    url='https://github.com/cubewise-code/tm1py',
    download_url=SCHEDULE_DOWNLOAD_URL,
    keywords=[
        'TM1', 'IBM Cognos TM1', 'Planning Analytics', 'PA', 'Cognos'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Natural Language :: English',
    ],
    install_requires=[
        'ijson',
        'requests',
        'pytz',
        'requests_negotiate_sspi;platform_system=="Windows"',
        'mdxpy>=0.4'],
    extras_require={
        "pandas": ["pandas"]
    },
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*-

import collections
import json
from typing import Optional, Iterable, Dict, List

from TM1py.Objects.Hierarchy import Hierarchy
from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils.Utils import case_and_space_insensitive_equals


class Dimension(TM1Object):
    """ Abstraction of TM1 Dimension

        A Dimension is a container for hierarchies.
    """

    def __init__(self, name: str, hierarchies: Optional[Iterable[Hierarchy]] = None):
        """ Abstraction of TM1 Dimension


        :param name: Name of the dimension
        :param hierarchies: List of TM1py.Objects.Hierarchy instances
        """
        self._name = name
        self._hierarchies = list(hierarchies) if hierarchies else []
        self._attributes = {'Caption': name}

    @classmethod
    def from_json(cls, dimension_as_json: str) -> 'Dimension':
        dimension_as_dict = json.loads(dimension_as_json)
        return cls.from_dict(dimension_as_dict)

    @classmethod
    def from_dict(cls, dimension_as_dict: Dict) -> 'Dimension':
        return cls(name=dimension_as_dict['Name'],
                   hierarchies=[Hierarchy.from_dict(hierarchy, dimension_name=dimension_as_dict['Name'])
                                for hierarchy
                                in dimension_as_dict['Hierarchies']])

    @property
    def name(self) -> str:
        return self._name

    @property
    def unique_name(self) -> str:
        return '[' + self._name + ']'

    @property
    def hierarchies(self) -> List[Hierarchy]:
        return self._hierarchies

    @property
    def hierarchy_names(self) -> List[str]:
        return [hierarchy.name for hierarchy in self._hierarchies]

    @property
    def default_hierarchy(self) -> Hierarchy:
        return self._hierarchies[0]

    @name.setter
    def name(self, value: str):
        for hierarchy in self.hierarchies:
            hierarchy._dimension_name = value
            if hierarchy.name == self._name:
                hierarchy.name = value
        self._name = value

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body())

    @property
    def body_as_dict(self) -> Dict:
        return self._construct_body()

    def __iter__(self):
        return iter(self._hierarchies)

    def __len__(self):
        return len(self.hierarchies)

    def __contains__(self, item):
        return self.contains_hierarchy(item)

    def __getitem__(self, item):
        return self.get_hierarchy(item)

    def contains_hierarchy(self, hierarchy_name: str) -> bool:
        for hierarchy in self._hierarchies:
            if case_and_space_insensitive_equals(hierarchy.name, hierarchy_name):
                return True
        return False

    def get_hierarchy(self, hierarchy_name: str) -> Hierarchy:
        for hierarchy in self._hierarchies:
            if case_and_space_insensitive_equals(hierarchy.name, hierarchy_name):
                return hierarchy
        raise ValueError("Hierarchy: {} not found in dimension: {}".format(hierarchy_name, self.name))

    def add_hierarchy(self, hierarchy: Hierarchy):
        if self.contains_hierarchy(hierarchy.name):
            raise ValueError("Hierarchy: {} already exists in dimension: {}".format(hierarchy.name, self.name))
        self._hierarchies.append(hierarchy)

    def remove_hierarchy(self, hierarchy_name: str):
        if case_and_space_insensitive_equals(hierarchy_name, "leaves"):
            raise ValueError("'Leaves' hierarchy must not be removed from dimension")

        for num, hierarchy in enumerate(self._hierarchies):
            if case_and_space_insensitive_equals(hierarchy.name, hierarchy_name):
                del self._hierarchies[num]
                return

    def _construct_body(self, include_leaves_hierarchy=False) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict["Name"] = self._name
        body_as_dict["UniqueName"] = self.unique_name
        body_as_dict["Attributes"] = self._attributes
        body_as_dict["Hierarchies"] = [
            hierarchy.body_as_dict
            for hierarchy
            in self.hierarchies if
            not case_and_space_insensitive_equals(hierarchy.name, "Leaves") or include_leaves_hierarchy]
        return body_as_dict



from abc import abstractmethod


class TM1Object:
    """ Parent Class for all TM1 Objects e.g. Cube, Process, Dimension.
    
    """
    SANDBOX_DIMENSION = "Sandboxes"

    @property
    @abstractmethod
    def body(self) -> str:
        pass

    def __hash__(self):
        return hash(self.body)

    def __str__(self):
        return self.body

    def __repr__(self):
        return "{}:{}".format(self.__class__.__name__, self.body)

    def __eq__(self, other):
        return self.body == other.body

    def __ne__(self, other):
        return self.body != other.body



from TM1py.Exceptions.Exceptions import TM1pyRestException, TM1pyException, TM1pyTimeout, TM1pyVersionException, \
    TM1pyNotAdminException, TM1pyWriteFailureException, TM1pyWritePartialFailureException



# -*- coding: utf-8 -*-

import json
from typing import List, Iterable

from requests import Response

from TM1py.Objects.User import User
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils.Utils import format_url, CaseAndSpaceInsensitiveSet, require_admin


class SecurityService(ObjectService):
    """ Service to handle Security stuff
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def determine_actual_user_name(self, user_name: str, **kwargs) -> str:
        return self.determine_actual_object_name(object_class="Users", object_name=user_name, **kwargs)

    def determine_actual_group_name(self, group_name: str, **kwargs) -> str:
        return self.determine_actual_object_name(object_class="Groups", object_name=group_name, **kwargs)

    @require_admin
    def create_user(self, user: User, **kwargs) -> Response:
        """ Create a user on TM1 Server

        :param user: instance of TM1py.User
        :return: response
        """
        url = '/api/v1/Users'
        return self._rest.POST(url, user.body, **kwargs)

    @require_admin
    def create_group(self, group_name: str, **kwargs) -> Response:
        """ Create a Security group in the TM1 Server

        :param group_name:
        :return:
        """
        url = '/api/v1/Groups'
        return self._rest.POST(url, json.dumps({"Name": group_name}), **kwargs)

    def get_user(self, user_name: str, **kwargs) -> User:
        """ Get user from TM1 Server

        :param user_name:
        :return: instance of TM1py.User
        """
        user_name = self.determine_actual_user_name(user_name, **kwargs)
        url = format_url(
            "/api/v1/Users('{}')?$select=Name,FriendlyName,Password,Type,Enabled&$expand=Groups",
            user_name)
        response = self._rest.GET(url, **kwargs)
        return User.from_dict(response.json())

    def get_current_user(self, **kwargs) -> User:
        """ Get user and group assignments of this session

        :return: instance of TM1py.User
        """
        url = "/api/v1/ActiveUser?$select=Name,FriendlyName,Password,Type,Enabled&$expand=Groups"
        response = self._rest.GET(url, **kwargs)
        return User.from_dict(response.json())

    @require_admin
    def update_user(self, user: User, **kwargs) -> Response:
        """ Update user on TM1 Server

        :param user: instance of TM1py.User
        :return: response
        """
        user.name = self.determine_actual_user_name(user.name, **kwargs)
        for current_group in self.get_groups(user.name, **kwargs):
            if current_group not in user.groups:
                self.remove_user_from_group(current_group, user.name, **kwargs)
        url = format_url("/api/v1/Users('{}')", user.name)
        return self._rest.PATCH(url, user.body, **kwargs)

    def update_user_password(self, user_name: str, password: str, **kwargs) -> Response:
        url = format_url("/api/v1/Users('{}')", user_name)
        body = {"Password": password}
        return self._rest.PATCH(url, json.dumps(body), **kwargs)

    @require_admin
    def delete_user(self, user_name: str, **kwargs) -> Response:
        """ Delete user on TM1 Server

        :param user_name:
        :return: response
        """
        user_name = self.determine_actual_user_name(user_name, **kwargs)
        url = format_url("/api/v1/Users('{}')", user_name)
        return self._rest.DELETE(url, **kwargs)

    @require_admin
    def delete_group(self, group_name: str, **kwargs) -> Response:
        """ Delete a group in the TM1 Server

        :param group_name:
        :return:
        """
        group_name = self.determine_actual_group_name(group_name, **kwargs)
        url = format_url("/api/v1/Groups('{}')", group_name)
        return self._rest.DELETE(url, **kwargs)

    def get_all_users(self, **kwargs):
        """ Get all users from TM1 Server

        :return: List of TM1py.User instances
        """
        url = '/api/v1/Users?$select=Name,FriendlyName,Password,Type,Enabled&$expand=Groups'
        response = self._rest.GET(url, **kwargs)
        users = [User.from_dict(user) for user in response.json()['value']]
        return users

    def get_all_user_names(self, **kwargs):
        """ Get all user names from TM1 Server

        :return: List of TM1py.User instances
        """
        url = '/api/v1/Users?select=Name'
        response = self._rest.GET(url, **kwargs)
        users = [user["Name"] for user in response.json()['value']]
        return users

    def get_users_from_group(self, group_name: str, **kwargs):
        """ Get all users from group

        :param group_name:
        :return: List of TM1py.User instances
        """
        url = format_url(
            "/api/v1/Groups('{}')?$expand=Users($select=Name,FriendlyName,Password,Type,Enabled;$expand=Groups)",
            group_name)
        response = self._rest.GET(url, **kwargs)
        users = [User.from_dict(user) for user in response.json()['Users']]
        return users

    def get_user_names_from_group(self, group_name: str, **kwargs) -> List[str]:
        """ Get all users from group

        :param group_name:
        :return: List of strings
        """
        url = format_url("/api/v1/Groups('{}')?$expand=Users($expand=Groups)", group_name)
        response = self._rest.GET(url, **kwargs)
        users = [user["Name"] for user in response.json()['Users']]
        return users

    def get_groups(self, user_name: str, **kwargs) -> List[str]:
        """ Get the groups of a user in TM1 Server

        :param user_name:
        :return: List of strings
        """
        user_name = self.determine_actual_user_name(user_name, **kwargs)
        url = format_url("/api/v1/Users('{}')/Groups", user_name)
        response = self._rest.GET(url, **kwargs)
        return [group['Name'] for group in response.json()['value']]

    @require_admin
    def add_user_to_groups(self, user_name: str, groups: Iterable[str], **kwargs) -> Response:
        """
        
        :param user_name: name of user
        :param groups: iterable of groups
        :return: response
        """
        user_name = self.determine_actual_user_name(user_name, **kwargs)
        url = format_url("/api/v1/Users('{}')", user_name)
        body = {
            "Name": user_name,
            "Groups@odata.bind": [
                format_url("Groups('{}')", self.determine_actual_group_name(group))
                for group
                in groups]
        }
        return self._rest.PATCH(url, json.dumps(body), **kwargs)

    @require_admin
    def remove_user_from_group(self, group_name: str, user_name: str, **kwargs) -> Response:
        """ Remove user from group in TM1 Server

        :param group_name:
        :param user_name:
        :return: response
        """
        user_name = self.determine_actual_user_name(user_name, **kwargs)
        group_name = self.determine_actual_group_name(group_name, **kwargs)
        url = format_url("/api/v1/Users('{}')/Groups?$id=Groups('{}')", user_name, group_name)
        return self._rest.DELETE(url, **kwargs)

    def get_all_groups(self, **kwargs) -> List[str]:
        """ Get all groups from TM1 Server

        :return: List of strings
        """
        url = '/api/v1/Groups?$select=Name'
        response = self._rest.GET(url, **kwargs)
        groups = [entry['Name'] for entry in response.json()['value']]
        return groups

    @require_admin
    def security_refresh(self, **kwargs) -> Response:
        from TM1py.Services import ProcessService
        ti = "SecurityRefresh;"
        process_service = ProcessService(self._rest)
        return process_service.execute_ti_code(ti, **kwargs)

    def user_exists(self, user_name: str, **kwargs) -> bool:
        url = format_url("/api/v1/Users('{}')", user_name)
        return self._exists(url, **kwargs)

    def group_exists(self, group_name: str, **kwargs) -> bool:
        url = format_url("/api/v1/Groups('{}')", group_name)
        return self._exists(url, **kwargs)

    def get_custom_security_groups(self, **kwargs) -> List[str]:
        custom_groups = CaseAndSpaceInsensitiveSet(*self.get_all_groups(**kwargs))
        custom_groups.discard('Admin')
        custom_groups.discard('DataAdmin')
        custom_groups.discard('SecurityAdmin')
        custom_groups.discard('OperationsAdmin')
        custom_groups.discard('}tp_Everyone')

        return list(custom_groups)

    def get_read_only_users(self, **kwargs) -> List[str]:
        read_only_users = list()

        mdx = """
        SELECT
        {[}ClientProperties].[ReadOnlyUser]} ON COLUMNS,
        NON EMPTY {[}Clients].MEMBERS} ON ROWS
        FROM [}ClientProperties]
        """

        from TM1py import CellService
        cell_service = CellService(self._rest)

        users_with_flag = cell_service.execute_mdx_rows_and_values(
            mdx=mdx,
            element_unique_names=False,
            **kwargs)

        for row, values in users_with_flag.items():
            user = row[0]
            read_only = values[0]
            if read_only:
                read_only_users.append(user)
        return read_only_users



# -*- coding: utf-8 -*-

import collections
import json
from typing import Dict, List, Iterable

from TM1py.Objects.ChoreFrequency import ChoreFrequency
from TM1py.Objects.ChoreStartTime import ChoreStartTime
from TM1py.Objects.ChoreTask import ChoreTask
from TM1py.Objects.TM1Object import TM1Object


class Chore(TM1Object):
    """ Abstraction of TM1 Chore

    """
    SINGLE_COMMIT = 'SingleCommit'
    MULTIPLE_COMMIT = 'MultipleCommit'

    def __init__(self, name: str, start_time: ChoreStartTime, dst_sensitivity: bool, active: bool,
                 execution_mode: str, frequency: ChoreFrequency, tasks: Iterable[ChoreTask]):
        self._name = name
        self._start_time = start_time
        self._dst_sensitivity = dst_sensitivity
        self._active = active
        self._execution_mode = execution_mode
        self._frequency = frequency
        self._tasks = list(tasks)

    @classmethod
    def from_json(cls, chore_as_json: str) -> 'Chore':
        """ Alternative constructor

        :param chore_as_json: string, JSON. Response of /api/v1/Chores('x')/Tasks?$expand=*
        :return: Chore, an instance of this class
        """
        chore_as_dict = json.loads(chore_as_json)
        return cls.from_dict(chore_as_dict)

    @classmethod
    def from_dict(cls, chore_as_dict: Dict) -> 'Chore':
        """ Alternative constructor

        :param chore_as_dict: Chore as dict
        :return: Chore, an instance of this class
        """
        return cls(name=chore_as_dict['Name'],
                   start_time=ChoreStartTime.from_string(chore_as_dict['StartTime']),
                   dst_sensitivity=chore_as_dict['DSTSensitive'],
                   active=chore_as_dict['Active'],
                   execution_mode=chore_as_dict['ExecutionMode'],
                   frequency=ChoreFrequency.from_string(chore_as_dict['Frequency']),
                   tasks=[ChoreTask.from_dict(task) for task in chore_as_dict['Tasks']])

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def start_time(self) -> ChoreStartTime:
        return self._start_time

    @start_time.setter
    def start_time(self, start_time: ChoreStartTime):
        self._start_time = start_time

    @property
    def dst_sensitivity(self) -> bool:
        return self._dst_sensitivity

    @dst_sensitivity.setter
    def dst_sensitivity(self, dst_sensitivity: bool):
        self._dst_sensitivity = dst_sensitivity

    @property
    def active(self) -> bool:
        return self._active

    @property
    def execution_mode(self) -> str:
        return self._execution_mode

    @execution_mode.setter
    def execution_mode(self, execution_mode):
        self._execution_mode = execution_mode

    @property
    def frequency(self) -> ChoreFrequency:
        return self._frequency

    @frequency.setter
    def frequency(self, frequency: ChoreFrequency):
        self._frequency = frequency

    @property
    def tasks(self) -> List[ChoreTask]:
        return self._tasks

    @tasks.setter
    def tasks(self, tasks: List[ChoreTask]):
        self._tasks = tasks

    @property
    def body(self) -> str:
        return self.construct_body()

    @property
    def body_as_dict(self) -> Dict:
        return json.loads(self.body)

    def add_task(self, task: ChoreTask):
        self._tasks.append(task)

    def activate(self):
        self._active = True

    def deactivate(self):
        self._active = False

    def reschedule(self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0):
        self._start_time.add(days=days, hours=hours, minutes=minutes, seconds=seconds)

    def construct_body(self) -> str:
        """
        construct self.body (json) from the class attributes
        :return: String, TM1 JSON representation of a chore
        """
        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self._name
        body_as_dict['StartTime'] = self._start_time.start_time_string
        body_as_dict['DSTSensitive'] = self._dst_sensitivity
        body_as_dict['Active'] = self._active
        body_as_dict['ExecutionMode'] = self._execution_mode
        body_as_dict['Frequency'] = self._frequency.frequency_string
        body_as_dict['Tasks'] = [task.body_as_dict for task in self._tasks]
        return json.dumps(body_as_dict, ensure_ascii=False)
